import importlib
import typing as t


def import_string(dotted_path: t.Any, package: str = None) -> t.Any:
    if not isinstance(dotted_path, str):
        return dotted_path

    # module import?
    if "." not in dotted_path:
        return importlib.import_module(dotted_path, package)

    module_path, attribute = dotted_path.rsplit(".", 1)
    module = importlib.import_module(module_path, package)
    try:
        return getattr(module, attribute)
    except AttributeError:
        raise ImportError(
            f'Module "{module_path}" does not have attribute "{attribute}"'
        )



import typing as t

from kupala.application import App
from kupala.authentication import LoginManager
from kupala.contracts import Authenticator, IdentityProvider, PasswordHasher
from kupala.extensions import Extension
from kupala.utils import import_string


class AuthenticatorConfig(t.TypedDict):
    authenticator: str
    user_provider: str
    options: t.Optional[dict[str, t.Any]]


class AuthenticationExtension(Extension):
    def __init__(
        self,
        user_providers: dict[str, IdentityProvider],
        default_provider: str = "default",
        authenticators: dict[str, AuthenticatorConfig] = None,
    ) -> None:
        self.default_provider = default_provider
        self.user_providers = user_providers
        self.authenticators = authenticators or {}

    def register(self, app: App) -> None:
        for name, user_provider in self.user_providers.items():
            app.bind(f"user_provider.{name}", user_provider)

        for name, config in self.authenticators.items():
            app.singleton(
                f"authenticator.{name}",
                self.create_authenticator(
                    config["authenticator"],
                    config.get("user_provider", "default"),
                    config.get("options", {}),
                ),
                tags=["authenticator"],
            )

        app.alias(f"user_provider.{self.default_provider}", IdentityProvider)
        app.singleton(
            LoginManager,
            self.create_login_manager,
            aliases=[
                "login_manager",
                "login_manager.default",
            ],
        )

    def create_authenticator(
        self,
        klass: str,
        user_provider: str,
        options: dict = None,
    ) -> t.Callable[[App], Authenticator]:
        def factory(app: App) -> Authenticator:
            options_ = options or {}
            provider = app.get(f"user_provider.{user_provider}")
            cls = import_string(klass)
            return cls(provider, **options_)

        return factory

    def create_login_manager(
        self,
        user_provider: IdentityProvider,
        hasher: PasswordHasher,
    ) -> LoginManager:
        return LoginManager(user_provider, hasher)



import asyncio
import typing as t

from starlette import exceptions
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala.requests import Request
from kupala.responses import Response, TemplateResponse, TextResponse

ErrorHandler = t.Callable[
    [Request, Exception],
    t.Union[Response, t.Awaitable[Response]],
]


def error_handler_factory(template: str) -> ErrorHandler:
    """Create a error handler callable for Jinja2 template."""

    def view(request: Request, exc: Exception) -> Response:
        return TemplateResponse(
            request,
            template,
            {
                "request": request,
                "exc": exc,
            },
        )

    return view


class ErrorHandlers:
    """A registry of errors handlers."""

    def __init__(self) -> None:
        self._handlers: dict[
            t.Union[int, t.Type[Exception]],
            ErrorHandler,
        ] = {}

    def use(
        self,
        code_or_exc: t.Union[int, t.Type[Exception]],
        handler: t.Union[ErrorHandler, str],
    ) -> None:
        """Add an error handler.
        You can pass either integer representing an HTTP status code
        or an exception class.
        The handler is a view function that accepts request
        and raised exception and returns a response."""
        if isinstance(handler, str):
            self._handlers[code_or_exc] = error_handler_factory(handler)
        else:
            self._handlers[code_or_exc] = handler

    def get(self, exc: Exception) -> t.Optional[ErrorHandler]:
        """Get an error handler for an exception."""
        if isinstance(exc, exceptions.HTTPException):
            handler = self._handlers.get(exc.__class__)
            if handler is None:
                handler = self._handlers.get(exc.status_code, self.http_handler)
            return handler

        handler = self._handlers.get(exc.__class__)
        if handler:
            return handler

        for cls in type(exc).__mro__:
            if cls in self._handlers:
                return self._handlers[cls]
        return None

    def http_handler(self, request: Request, exc: Exception) -> Response:
        """For HTTP errors we display a simple text message
        and set a correct HTTP response status."""
        assert isinstance(exc, exceptions.HTTPException)
        if exc.status_code in {204, 304}:
            return Response(b"", status_code=exc.status_code)
        return TextResponse(exc.detail, exc.status_code)


class ExceptionMiddleware:
    def __init__(self, app: ASGIApp, handlers: ErrorHandlers):
        self.app = app
        self.handlers = handlers

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: nocover
            return await self.app(scope, receive, send)

        response_started = False

        async def sender(message: Message) -> None:
            nonlocal response_started
            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, sender)
        except Exception as exc:
            handler = self.handlers.get(exc)
            if handler is None:
                raise exc from None

            if response_started:
                msg = "Caught handled exception, but response already started."
                raise RuntimeError(msg) from exc

            request = scope["request"]
            if asyncio.iscoroutinefunction(handler):
                # todo: fix typing error
                response = await handler(request, exc)  # type: ignore
            else:
                response = await asyncio.to_thread(handler, request, exc)
            await response(scope, receive, sender)



import typing as t
from dataclasses import dataclass, field

if t.TYPE_CHECKING:
    from kupala.application import App


class Extension(t.Protocol):
    def register(self, app: "App") -> None:
        pass

    def bootstrap(self, app: "App") -> None:
        pass


class BaseExtension:
    def register(self, app: "App") -> None:
        pass

    def bootstrap(self, app: "App") -> None:
        pass


@dataclass
class Extensions:
    _extensions: list[Extension] = field(default_factory=list)

    def use(self, extension: Extension) -> None:
        self._extensions.append(extension)

    def __iter__(self) -> t.Iterator[Extension]:
        return iter(self._extensions)



from kupala.application import App
from kupala.contracts import URLResolver
from kupala.extensions import Extension
from kupala.routing import Router, Routes, RouteURLResolver


class RoutingExtension(Extension):
    def __init__(self, redirect_slashes: bool = True) -> None:
        self.redirect_slashes = redirect_slashes

    def register(self, app: App) -> None:
        app.singleton(Router, self._create_router, aliases="router")
        app.singleton(
            URLResolver,
            self._create_url_resolver,
            aliases=[
                "url_resolver",
                "route_url_resolver",
                RouteURLResolver,
            ],
        )

    def _create_router(self, app: App) -> Router:
        return Router(
            routes=app.get(Routes),
            redirect_slashes=self.redirect_slashes,
            lifespan=app.lifespan,
        )

    def _create_url_resolver(self, router: Router) -> RouteURLResolver:
        return RouteURLResolver(router)



import builtins
import functools
import os
import typing as t

import dotenv as dotenvlib

from . import json


def noop_coerce(x: t.Any) -> None:
    return x


class DotenvError(Exception):
    pass


def list_(
    value: t.Union[str, t.Iterable],
    coerce: t.Callable[[str], t.Any] = noop_coerce,
    split_char: str = " ",
) -> list[t.Any]:
    if isinstance(value, str):
        values = map(str.strip, value.split(split_char))
        return list(
            map(
                coerce,
                map(
                    lambda x: x.strip("\"'"),
                    values,
                ),
            )
        )
    return list(value)


class Env:
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix
        self._values = t.cast(dict[str, t.Optional[str]], os.environ.copy())

    def load(self, dot_file: str) -> str:
        """Load .env file."""
        dot_path = dotenvlib.find_dotenv(str(dot_file))
        self._import(dotenvlib.dotenv_values(dot_path))
        return dot_path

    def _import(self, values: t.Mapping[str, t.Optional[str]]) -> None:
        for key, value in values.items():
            if key.startswith(self.prefix):
                offset = 0
                if len(self.prefix):
                    offset = 1
                self._values[key[len(self.prefix) + offset :]] = value

    def all(self) -> dict:
        """Return all environmental value."""
        return self._values

    def str(self, name: str, default: str = None) -> str:
        """Get a variable as a string."""
        return self.get(name, default, str)

    def bytes(
        self,
        name: builtins.str,
        default: bytes = None,
        encoding: builtins.str = "utf8",
    ) -> bytes:
        """Get a variable as bytes."""
        value = self.get(name, default)
        if hasattr(value, "encode"):
            value = value.encode(encoding)
        return bytes(value)

    def int(self, name: builtins.str, default: int = None) -> int:
        """Get a variable as an integer."""
        return self.get(name, default, int)

    def float(self, name: builtins.str, default: float = None) -> float:
        """Get a variable as a float."""
        return self.get(name, default, float)

    def bool(
        self,
        name: builtins.str,
        default: bool = None,
    ) -> t.Optional[bool]:
        """Get a variable as a boolean value."""
        value = self.get(name, None)
        if value is None:
            return default
        return value.lower() in ["yes", "y", "1"]

    def list(
        self,
        name: builtins.str,
        default: list[t.Any] = None,
        coerce: t.Callable[[builtins.str], t.Any] = noop_coerce,
        split_char: builtins.str = ",",
    ) -> builtins.list[t.Any]:
        """Get a list from the environment variables.
        Use `coerce` function to cast each item to a specific type.
        You can set a different separator using `split_char` argument."""
        caster = functools.partial(list_, coerce=coerce, split_char=split_char)
        return self.get(name, default, caster)

    def csv(
        self,
        name: builtins.str,
        default: builtins.list[t.Any] = None,
        coerce: t.Callable[[builtins.str], t.Any] = noop_coerce,
        delimiter: builtins.str = ",",
    ) -> builtins.list[t.Any]:
        """Parse a CSV string into list from the environment variable.
        Use `coerce` function to cast each item to a specific type."""
        caster = functools.partial(list_, coerce=coerce, split_char=delimiter)
        return self.get(name, default, caster)

    def json(self, name: builtins.str, default: t.Any = None) -> t.Any:
        """Decode JSON string from the environment variable."""
        value = self.get(name, default)
        if isinstance(value, str):
            return json.loads(value)
        return default

    def get(
        self,
        name: builtins.str,
        default: t.Any = None,
        cast: t.Callable[[t.Any], t.Any] = None,
    ) -> t.Any:
        """Get an environment variable.
        Use `cast` callable to convert a value to a specific type."""
        value = self._values.get(name, default)
        if value is not None and cast:
            value = cast(value)
        return value

    def set(self, name: builtins.str, value: builtins.str) -> None:
        """Set an environment variable.
        Note, it won't set a global variable
        but will update local variable cache."""
        self._values[name] = str(value)

    def delete(self, name: builtins.str) -> None:
        """Delete a variable from the local cache."""
        del self._values[name]

    __call__ = str
    __getitem__ = get
    __setitem__ = set
    __delitem__ = delete

    def __repr__(self) -> builtins.str:
        return "<Env: %s>" % {k: v for k, v in self.all().items()}

    def __contains__(self, name: builtins.str) -> builtins.bool:
        return name in self._values



import typing as t
from dataclasses import dataclass

from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.contracts import Authenticator, Identity, IdentityProvider, PasswordHasher
from kupala.requests import Request

SESSION_KEY = "_user"


class AuthenticationError(Exception):
    """Base error for authentication exceptions."""


@dataclass
class AuthState:
    """Contains authentication state."""

    user: t.Optional[Identity] = None

    @property
    def is_anonymous(self) -> bool:
        return not self.is_authenticated

    @property
    def is_authenticated(self) -> bool:
        return self.user is not None

    def clear(self) -> None:
        self.user = None

    def __bool__(self) -> bool:
        return self.is_authenticated


class LoginManager:
    """Log in or log our users."""

    def __init__(
        self,
        user_provider: IdentityProvider,
        hasher: PasswordHasher = None,
    ) -> None:
        self.user_provider = user_provider
        self.hasher = hasher

    async def login(
        self, request: Request, identity: str, credential: str
    ) -> t.Optional[Identity]:
        user = await self.user_provider.find_by_identity(identity)
        if user:
            hasher = self.hasher or request.app.get(PasswordHasher)
            hashed = user.get_hashed_password()
            if hasher.verify(credential, hashed):
                request.session[SESSION_KEY] = user.get_id()
                await request.session.regenerate_id()
                return user
        return None

    async def logout(self, request: Request) -> None:
        request.session.clear()
        await request.session.regenerate_id()
        request.auth.clear()


class SessionAuthenticator:
    """Session authenticator uses session key as a source of user's identifier.
    Once found, it attempts to load user using found ID."""

    def __init__(self, user_provider: IdentityProvider):
        self.user_provider = user_provider

    async def authenticate(self, request: Request) -> t.Optional[Identity]:
        if "session" not in request.scope:  # pragma: nocover
            raise KeyError("AuthenticationMiddleware requires SessionMiddleware.")
        user_id = request.session.get(SESSION_KEY)
        if not user_id:
            return None
        return await self.user_provider.find_by_id(user_id)


class TokenAuthenticator:
    """Token authenticator reads a header of specific type from
    the request and loads user using token from the header.

    Default header name is 'Authorization' and token type is 'Bearer'."""

    header_name = "Authorization"

    def __init__(self, user_provider: IdentityProvider, token: str = "Bearer"):
        self.user_provider = user_provider
        self.token = token

    async def authenticate(self, request: Request) -> t.Optional[Identity]:
        header = request.headers.get(self.header_name)
        if header:
            token_type, value = header.split(" ")
            if token_type == self.token:
                return await self.user_provider.find_by_token(value)
        return None


class AuthenticationMiddleware:
    """The middleware looks up for a user using list of authenticators.
    The first authenticator that returns Identity instance will be considered
    as successful and will break the iteration.

    You can access loaded identity using `request.auth` property.
    If the user wasn't loaded the `request.auth` still available but is not
    authenticated.

    You can check for user authentication by testing `request.auth`:

        def view(request):
            if request.auth:
                print('authenticated')
            else:
                print('not authenticated')
    """

    def __init__(
        self,
        app: ASGIApp,
        authenticators: list[t.Union[Authenticator, str]] = None,
    ) -> None:
        self.app = app
        self.authenticators = authenticators or []

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        scope["auth"] = AuthState()
        request: Request = scope["request"]
        for authenticator in self.authenticators:
            if isinstance(authenticator, str):
                authenticator = f"authenticator.{authenticator}"
            authenticator = t.cast(Authenticator, request.app.get(authenticator))
            user = await authenticator.authenticate(request)
            if user:
                scope["auth"] = AuthState(user)
                break
        await self.app(scope, receive, send)


class InMemoryProvider(IdentityProvider):
    """A simple user provider that uses a dictionary as the source of users."""

    def __init__(self, users: list[Identity]):
        self.users = users

    @property
    def by_id(self) -> dict[str, Identity]:
        return {u.get_id(): u for u in self.users}

    @property
    def by_identity(self) -> dict[str, Identity]:
        return {u.get_identity(): u for u in self.users}

    async def find_by_identity(self, identity: str) -> t.Optional[Identity]:
        return self.by_identity.get(identity)

    async def find_by_id(self, id: t.Any) -> t.Optional[Identity]:
        return self.by_id.get(id)

    async def find_by_token(self, token: str) -> t.Optional[Identity]:
        return self.by_identity.get(token)



from kupala.application import App
from kupala.contracts import PasswordHasher
from kupala.extensions import Extension
from kupala.utils import import_string


class HashingExtension(Extension):
    default_hashers = {
        "insecure": "kupala.security.hashers.InsecureHasher",
        "pbkdf2_sha256": "kupala.security.hashers.Pbkdf2Sha256Hasher",
        "pbkdf2_sha512": "kupala.security.hashers.Pbkdf2Sha512Hasher",
    }

    def __init__(
        self,
        hashers: dict[str, str] = None,
        default: str = "pbkdf2_sha256",
    ):
        self.default = default
        self.hashers = self.default_hashers
        if hashers:
            self.hashers.update(hashers)

    def register(self, app: App) -> None:
        for name, hasher in self.hashers.items():
            hasher_class = import_string(hasher)
            app.singleton(
                hasher_class,
                hasher_class,
                aliases=f"hasher.{name}",
                tags=["password_hasher"],
            )

        app.alias(f"hasher.{self.default}", PasswordHasher)



from __future__ import annotations

import typing as t

from starlette.middleware import base
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.requests import Request
from kupala.responses import Response


class Middleware:
    """Keeps middleware callable along with its constructor arguments."""

    def __init__(self, obj: t.Type, **kwargs: t.Any) -> None:
        self.obj = obj
        self.args = kwargs

    def wrap(self, app: ASGIApp) -> ASGIApp:
        """Wraps ASGI application with this middleware."""
        return self.obj(app, **self.args)

    def __repr__(self) -> str:  # pragma: no cover
        return "<Middleware: %s, kwargs=%r>" % (self.obj, self.args)


class MiddlewareStack:
    """Keeps track about all middleware used."""

    def __init__(self) -> None:
        self._global: list[Middleware] = []
        self._groups: dict[str, MiddlewareStack] = {}

    def top(self, mw: t.Type, **kwargs: t.Any) -> None:
        """Add middleware to the top of stack."""
        self._global.insert(0, Middleware(mw, **kwargs))

    def use(self, mw: t.Type, **kwargs: t.Any) -> None:
        """Add middleware to the end of stack."""
        self._global.append(Middleware(mw, **kwargs))

    def group(self, name: str) -> MiddlewareStack:
        """Create a middleware group."""
        self._groups.setdefault(name, MiddlewareStack())
        return self._groups[name]

    def has_group(self, name: str) -> bool:
        return name in self._groups

    def __enter__(self) -> MiddlewareStack:
        return self

    def __exit__(self, *args: t.Any) -> None:
        pass

    def __iter__(self) -> t.Iterator[Middleware]:
        return iter(self._global)

    def __reversed__(self) -> t.Iterator[Middleware]:
        return reversed(self._global)

    def __len__(self) -> int:
        return len(self._global)

    def __str__(self) -> str:  # pragma: nocover
        return "<MiddlewareStack: %s middleware, %s groups>" % (
            len(self),
            len(self._groups.keys()),
        )


RequestResponseEndpoint = t.Callable[[Request], t.Awaitable[Response]]


class BaseHTTPMiddleware(base.BaseHTTPMiddleware):
    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        response = await self.dispatch_func(scope["request"], self.call_next)
        await response(scope, receive, send)

    async def dispatch(  # type: ignore[override]
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        raise NotImplementedError()  # pragma: no cover



from starlette import exceptions


class HttpException(exceptions.HTTPException):
    """Base exception for HTTP related errors."""

    status_code: int

    def __init__(self, message: str = None) -> None:  # pragma: nocover
        super().__init__(self.status_code, message)


class NotAuthenticatedError(HttpException):
    status_code = 401


class NotAuthorizedError(HttpException):
    status_code = 403


class NotFoundError(HttpException):
    status_code = 404


class MethodNotAllowedError(HttpException):
    status_code = 405


class ConflictError(HttpException):
    status_code = 409



import abc
import typing as t
import uuid
from base64 import b64decode, b64encode

from itsdangerous import BadSignature, SignatureExpired, TimestampSigner
from starlette.datastructures import MutableHeaders, Secret
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala import json


class SessionError(Exception):
    ...


class SessionNotLoaded(SessionError):
    pass


class SessionBackend(abc.ABC):
    """Base class for session backends."""

    @abc.abstractmethod
    async def read(self, session_id: str) -> t.Dict[str, t.Any]:  # pragma: no cover
        """Read session data from the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def write(
        self, data: t.Dict, session_id: t.Optional[str] = None
    ) -> str:  # pragma: no cover
        """Write session data to the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def remove(self, session_id: str) -> None:  # pragma: no cover
        """Remove session data from the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def exists(self, session_id: str) -> bool:  # pragma: no cover
        """Test if storage contains session data for a given session_id."""
        raise NotImplementedError()

    async def generate_id(self) -> str:
        """Generate a new session id."""
        return str(uuid.uuid4())


class CookieBackend(SessionBackend):
    """Stores session data in the browser's cookie as a signed string."""

    def __init__(self, secret_key: t.Union[str, Secret], max_age: int):
        self._signer = TimestampSigner(str(secret_key))
        self._max_age = max_age

    async def read(self, session_id: str) -> t.Dict:
        """A session_id is a signed session value."""
        try:
            data = self._signer.unsign(session_id, max_age=self._max_age)
            return json.loads(b64decode(data).decode())
        except (BadSignature, SignatureExpired):
            return {}

    async def write(self, data: t.Dict, session_id: t.Optional[str] = None) -> str:
        """The data is a session id in this backend."""
        encoded_data = b64encode(json.dumps(data).encode("utf-8"))
        return self._signer.sign(encoded_data).decode("utf-8")

    async def remove(self, session_id: str) -> None:
        """Session data stored on client side - no way to remove it."""

    async def exists(self, session_id: str) -> bool:
        return False


class InMemoryBackend(SessionBackend):
    """Stores session data in a dictionary."""

    def __init__(self, data: dict = None) -> None:
        self.data: dict = data or {}

    async def read(self, session_id: str) -> t.Dict:
        return self.data.get(session_id, {}).copy()

    async def write(self, data: t.Dict, session_id: t.Optional[str] = None) -> str:
        session_id = session_id or await self.generate_id()
        self.data[session_id] = data
        return session_id

    async def remove(self, session_id: str) -> None:
        del self.data[session_id]

    async def exists(self, session_id: str) -> bool:
        return session_id in self.data


class Session:
    def __init__(self, backend: SessionBackend, session_id: str = None) -> None:
        self.session_id = session_id
        self._data: dict[str, t.Any] = {}
        self._backend = backend
        self.is_loaded = False
        self._is_modified = False

    @property
    def is_empty(self) -> bool:
        """Check if session has data."""
        return len(self.keys()) == 0

    @property
    def is_modified(self) -> bool:
        """Check if session data has been modified,"""
        return self._is_modified

    @property
    def data(self) -> dict:
        if not self.is_loaded:
            raise SessionNotLoaded("Attempt to read from not loaded session.")
        return self._data

    @data.setter
    def data(self, value: dict[str, t.Any]) -> None:
        self._data = value

    async def load(self) -> None:
        """Load data from the backend.
        Subsequent calls do not take any effect."""
        if self.is_loaded:
            return

        if not self.session_id:
            self.data = {}
        else:
            self.data = await self._backend.read(self.session_id)

        self.is_loaded = True

    async def persist(self) -> str:
        self.session_id = await self._backend.write(self.data, self.session_id)
        return self.session_id

    async def delete(self) -> None:
        if self.session_id:
            self.data = {}
            self._is_modified = True
            await self._backend.remove(self.session_id)

    async def flush(self) -> str:
        self._is_modified = True
        await self.delete()
        return await self.regenerate_id()

    async def regenerate_id(self) -> str:
        self.session_id = await self._backend.generate_id()
        self._is_modified = True
        return self.session_id

    def keys(self) -> t.KeysView[str]:
        return self.data.keys()

    def values(self) -> t.ValuesView[t.Any]:
        return self.data.values()

    def items(self) -> t.ItemsView[str, t.Any]:
        return self.data.items()

    def pop(self, key: str, default: t.Any = None) -> t.Any:
        self._is_modified = True
        return self.data.pop(key, default)

    def get(self, name: str, default: t.Any = None) -> t.Any:
        return self.data.get(name, default)

    def setdefault(self, key: str, default: t.Any) -> None:
        self._is_modified = True
        self.data.setdefault(key, default)

    def clear(self) -> None:
        self._is_modified = True
        self.data.clear()

    def update(self, *args: t.Any, **kwargs: t.Any) -> None:
        self._is_modified = True
        self.data.update(*args, **kwargs)

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __setitem__(self, key: str, value: t.Any) -> None:
        self._is_modified = True
        self.data[key] = value

    def __getitem__(self, key: str) -> t.Any:
        return self.data[key]

    def __delitem__(self, key: str) -> None:
        self._is_modified = True
        del self.data[key]


class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        secret_key: t.Union[str, Secret],
        session_cookie: str = "session",
        max_age: int = 14 * 24 * 60 * 60,  # 14 days, in seconds
        same_site: str = "lax",
        https_only: bool = False,
        backend: SessionBackend = None,
    ) -> None:
        self.app = app
        self.backend = backend or CookieBackend(secret_key, max_age)
        self.session_cookie = session_cookie
        self.max_age = max_age
        self.security_flags = "httponly; samesite=" + same_site
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        session_id = connection.cookies.get(self.session_cookie, None)
        scope["session"] = Session(self.backend, session_id)

        async def send_wrapper(message: Message) -> None:
            if message["type"] == "http.response.start":
                session: Session = scope["session"]
                if session.is_modified and not session.is_empty:
                    # We have session data to persist (data was changed, cleared, etc).
                    nonlocal session_id
                    session_id = await scope["session"].persist()

                    headers = MutableHeaders(scope=message)
                    header_value = "%s=%s; path=/; Max-Age=%d; %s" % (
                        self.session_cookie,
                        session_id,
                        self.max_age,
                        self.security_flags,
                    )
                    headers.append("Set-Cookie", header_value)
                elif session.is_loaded and session.is_empty:
                    # no interactions to session were done
                    headers = MutableHeaders(scope=message)
                    header_value = "%s=%s; %s" % (
                        self.session_cookie,
                        "null; path=/; expires=Thu, 01 Jan 1970 00:00:00 GMT;",
                        self.security_flags,
                    )
                    headers.append("Set-Cookie", header_value)
            await send(message)

        await scope["session"].load()
        await self.app(scope, receive, send_wrapper)



import abc
import typing as t

from kupala.requests import Request

Debug = bool


@t.runtime_checkable
class Identity(t.Protocol):  # pragma: no cover
    def get_identity(self) -> str:
        ...

    def get_id(self) -> t.Any:
        ...

    def get_hashed_password(self) -> str:
        ...


class IdentityProvider(abc.ABC):  # pragma: no cover
    async def find_by_identity(self, identity: str) -> t.Optional[Identity]:
        raise NotImplementedError()

    async def find_by_id(self, id: t.Any) -> t.Optional[Identity]:
        raise NotImplementedError()

    async def find_by_token(self, token: str) -> t.Optional[Identity]:
        raise NotImplementedError()


class TemplateRenderer(t.Protocol):  # pragma: no cover
    """A protocol that all renderer implementations must match."""

    def render(self, template: str, context: dict[str, t.Any] = None) -> str:
        ...


class URLResolver(abc.ABC):  # pragma: no cover
    """URL resolver specification."""

    @abc.abstractmethod
    def resolve(
        self,
        name: str,
        **path_params: dict,
    ) -> str:  # pragma: nocover
        raise NotImplementedError()


class Authenticator(t.Protocol):
    def __init__(self, user_provider: IdentityProvider, **kwargs: t.Any):
        ...

    async def authenticate(self, request: Request) -> t.Optional[Identity]:
        ...


class PasswordHasher(t.Protocol):
    def hash(self, plain: str) -> str:
        ...

    def verify(self, plain: str, hashed: str) -> bool:
        ...



import os
import typing as t
from pathlib import Path
from urllib.parse import quote_plus

from starlette import responses
from starlette.background import BackgroundTask
from starlette.types import Receive, Scope, Send

from . import json
from .contracts import TemplateRenderer, URLResolver
from .requests import Request


class Response(responses.Response):
    ...


class TextResponse(responses.PlainTextResponse, Response):
    ...


class HTMLResponse(responses.HTMLResponse):
    ...


class FileResponse(responses.FileResponse):
    def __init__(
        self,
        path: t.Union[str, Path],
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        file_name: str = None,
        inline: bool = False,
    ):
        headers = headers or {}
        file_name = file_name or os.path.basename(path)
        encoded = quote_plus(file_name)
        disposition = "inline" if inline else "attachment"
        headers["Content-Disposition"] = f'{disposition}; filename="{encoded}"'
        super().__init__(
            path,
            status_code,
            headers,
            media_type=media_type,
        )


class StreamingResponse(responses.StreamingResponse):
    def __init__(
        self,
        content: t.Any,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        file_name: str = None,
        inline: bool = False,
    ):
        disposition = "inline" if inline else "attachment"
        headers = headers or {}
        if file_name:
            headers.update(
                {"content-disposition": f'{disposition}; filename="{file_name}"'}
            )
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
        )


class RedirectResponse(responses.RedirectResponse):
    def __init__(
        self,
        request: Request,
        url: str,
        status_code: int = 302,
        headers: dict = None,
        data: t.Any = None,
    ):
        if "session" in request.scope and data:
            request.session["_redirect_data"] = data

        if not url.startswith("/") and not url.startswith("http"):
            url = request.app.get(URLResolver).resolve(url)

        super().__init__(url, status_code, headers)


class JSONResponse(responses.Response):
    def __init__(
        self,
        content: t.Any,
        status_code: int = 200,
        headers: dict = None,
        indent: int = None,
        default: t.Callable = None,
        encoder_class: t.Type[json.JSONEncoder] = None,
        allow_nan: bool = True,
    ):
        content = json.dumps(
            content,
            cls=encoder_class,
            default=default,
            indent=indent,
            allow_nan=allow_nan,
        )
        super().__init__(content, status_code, headers, "application/json")


class TemplateResponse(Response):
    """A template response.

    This response renders a template using a configured renderer."""

    media_type = "text/html"

    def __init__(
        self,
        request: Request,
        template: str,
        context: dict = None,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ):
        self.request = request
        self.renderer = request.app.get(TemplateRenderer)
        self.template = template
        self.context = context or {}
        if "request" not in self.context:
            self.context["request"] = request

        content = self.renderer.render(template, self.context)
        super().__init__(content, status_code, headers, media_type, background)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        extensions = self.request.get("extensions", {})
        if "http.response.template" in extensions:
            await send(
                {
                    "type": "http.response.template",
                    "template": self.template,
                    "context": self.context,
                }
            )
        await super().__call__(scope, receive, send)


class ResponseFactory:
    def __init__(
        self, request: "Request", status_code: int = 200, headers: dict = None
    ) -> None:
        self.request = request
        self.status_code = status_code
        self.headers = headers

    def template(self, template_name: str, context: dict = None) -> TemplateResponse:
        return TemplateResponse(
            self.request,
            template_name,
            context,
            status_code=self.status_code,
            headers=self.headers,
        )

    def json(
        self,
        data: dict,
        default: t.Callable = None,
        indent: int = None,
        encoder_class: t.Type[json.JSONEncoder] = None,
    ) -> JSONResponse:
        return JSONResponse(
            data,
            status_code=self.status_code,
            headers=self.headers,
            default=default,
            indent=indent,
            encoder_class=encoder_class,
        )

    def text(self, content: str) -> TextResponse:
        return TextResponse(
            content,
            status_code=self.status_code,
            headers=self.headers,
        )

    def html(self, content: str) -> HTMLResponse:
        return HTMLResponse(
            content,
            status_code=self.status_code,
            headers=self.headers,
        )

    def stream(
        self,
        content: t.Any,
        file_name: str = "data.bin",
        media_type: str = "application/octet-stream",
        inline: bool = False,
    ) -> StreamingResponse:
        return StreamingResponse(
            content=content,
            status_code=self.status_code,
            headers=self.headers,
            file_name=file_name,
            inline=inline,
            media_type=media_type,
        )

    def redirect(
        self,
        url: str,
        data: t.Any = None,
        status_code: int = 302,
    ) -> RedirectResponse:
        return RedirectResponse(
            self.request,
            url=url,
            status_code=status_code or self.status_code,
            headers=self.headers,
            data=data,
        )

    def back(
        self,
        data: t.Any = None,
        status_code: int = 302,
    ) -> RedirectResponse:
        redirect_to = self.request.headers.get("referer", "/")
        current_origin = self.request.url.netloc
        if current_origin not in redirect_to:
            redirect_to = "/"
        return self.redirect(
            url=redirect_to,
            status_code=status_code,
            data=data,
        )

    def send_file(
        self,
        path: t.Union[str, Path],
        file_name: str,
        content_type: str = "application/octet-stream",
        inline: bool = False,
    ) -> FileResponse:
        return FileResponse(
            path=path,
            status_code=self.status_code,
            headers=self.headers,
            media_type=content_type,
            file_name=file_name,
            inline=inline,
        )


def response(
    request: Request,
    status_code: int = 200,
    headers: dict = None,
) -> ResponseFactory:
    return ResponseFactory(request, status_code, headers)



from __future__ import annotations

import pathlib
import typing as t
from unittest import mock

from starlette import routing
from starlette.concurrency import run_in_threadpool
from starlette.datastructures import URLPath
from starlette.routing import Match, iscoroutinefunction_or_partial
from starlette.staticfiles import StaticFiles
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.middleware import MiddlewareStack
from kupala.utils import import_string

from . import responses


def request_response(func: t.Callable) -> ASGIApp:
    """
    Takes a function or coroutine `func(request) -> response`,
    and returns an ASGI application.
    """
    is_coroutine = iscoroutinefunction_or_partial(func)

    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request = scope["request"]
        if is_coroutine:
            response = await request.app.invoke(func, request=request)
        else:
            response = await run_in_threadpool(
                request.app.invoke,
                func,
                request=request,
            )

        await response(scope, receive, send)

    return app


class Router(routing.Router):
    ...


class Route(routing.Route):
    def __init__(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        methods: list[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> None:
        with mock.patch("starlette.routing.request_response", request_response):
            super().__init__(
                path,
                endpoint,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
            )


class Mount(routing.Mount):
    ...


class WebSocketRoute(routing.WebSocketRoute):
    ...


class Redirect:
    def __init__(self, destination: str, status_code: int = 302) -> None:
        self.redirect_slashes = False
        self.destination = destination
        self.status_code = status_code

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        response = responses.RedirectResponse(
            scope["request"],
            self.destination,
            self.status_code,
        )
        await response(scope, receive, send)


class Host(routing.BaseRoute):
    def __init__(self, host: str, middleware: list[str] = None):
        self.host = host
        self.middleware = middleware or []
        self.routes = Routes()
        self.host_app: t.Optional[routing.Host] = None
        self.host_app_with_middleware: t.Optional[ASGIApp] = None

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        host_app = t.cast(ASGIApp, self._get_host_app())
        if self.host_app_with_middleware is None:
            stack: MiddlewareStack = scope["app"].middleware
            for group in self.middleware:
                if not stack.has_group(group):
                    raise KeyError('Middleware group "%s" is not defined.' % (group,))
                for middleware in reversed(stack.group(group)):
                    host_app = middleware.wrap(host_app)
            self.host_app_with_middleware = host_app
        await self.host_app_with_middleware(scope, receive, send)

    def matches(self, scope: Scope) -> t.Tuple[Match, Scope]:
        return self._get_host_app().matches(scope)

    def url_path_for(self, name: str, **path_params: str) -> URLPath:
        return self._get_host_app().url_path_for(name, **path_params)

    def _get_host_app(self) -> routing.Host:
        if self.host_app is None:
            self.host_app = routing.Host(self.host, Router(self.routes))
        return self.host_app

    def __enter__(self) -> Routes:
        return self.routes

    def __exit__(self, *args: t.Any) -> None:
        pass


class Group:
    """Group multiple similar routes under a common path.

    The routes can be wrapped into middleware.
    These middleware are defined on application level
    and grouped by a string key.
    """

    def __init__(self, middleware_groups: list[str] = None) -> None:
        self.middleware_groups = middleware_groups or []
        self.routes = Routes()
        self.wrapped_app: t.Optional[ASGIApp] = None

    def __enter__(self) -> Routes:
        return self.routes

    def __exit__(self, *args: t.Any) -> None:
        pass

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        if self.wrapped_app is None:
            self.wrapped_app = Router(self.routes)
            stack: MiddlewareStack = scope["app"].middleware
            for group in self.middleware_groups:
                for middleware in reversed(stack.group(group)):
                    self.wrapped_app = middleware.wrap(self.wrapped_app)

        assert self.wrapped_app
        return await self.wrapped_app(scope, receive, send)


class Routes(t.Sequence[routing.BaseRoute]):
    def __init__(self) -> None:
        self._routes: list[routing.BaseRoute] = []

    def get(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to GET requests."""
        self.add(path, endpoint, ["GET"], name)

    def post(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to POST requests."""
        self.add(path, endpoint, ["POST"], name)

    def put(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to PUT requests."""
        self.add(path, endpoint, ["PUT"], name)

    def patch(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to PATCH requests."""
        self.add(path, endpoint, ["PATCH"], name)

    def delete(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to DELETE requests."""
        self.add(path, endpoint, ["DELETE"], name)

    def options(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to OPTIONS requests."""
        self.add(path, endpoint, ["OPTIONS"], name)

    def any(self, path: str, endpoint: t.Callable, name: str = None) -> None:
        """Make endpoint to respond to any request method."""
        _all = ["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"]
        self.add(path, endpoint, _all, name)

    def get_or_post(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Make endpoint to respond to POST or GET requests."""
        self.add(path, endpoint, ["POST", "GET"], name)

    def add(
        self,
        path: str,
        endpoint: t.Callable,
        methods: t.Optional[list[str]] = None,
        name: str = None,
    ) -> None:
        """Add an endpoint."""
        self._routes.append(Route(path, endpoint=endpoint, methods=methods, name=name))

    def websocket(
        self,
        path: str,
        endpoint: t.Callable,
        name: str = None,
    ) -> None:
        """Add websocket endpoint."""
        self._routes.append(WebSocketRoute(path, endpoint, name=name))

    def group(
        self,
        path: str,
        middleware: list[str] = None,
    ) -> Group:
        """Group groups by a common path.
        Optionally, apply a list of middleware on this group.
        These middleware will be called before calling the endpoint."""
        group = Group(middleware)
        self.mount(path, group)
        return group

    def redirect(
        self,
        from_: str,
        destination: str,
        status_code: int = 302,
    ) -> None:
        """Redirect client from one route to another."""
        self.mount(
            from_,
            Redirect(destination=destination, status_code=status_code),
        )

    def host(
        self,
        domain: str,
        middleware: list[str] = None,
    ) -> Host:
        """Group groups by a subdomain.

        Optionally, apply a list of middleware on this group.
        These middleware will be called before calling the endpoint."""
        subdomain = Host(domain, middleware)
        self._routes.append(subdomain)
        return subdomain

    def mount(self, path: str, app: ASGIApp, name: str = None) -> None:
        """Mount any ASGI compliant object."""
        self._routes.append(Mount(path, app=app, name=name))

    def static(self, path: str, directory: t.Union[str, pathlib.Path]) -> None:
        self.mount(path, StaticFiles(directory=str(directory)))

    def include(self, module_name: str, callback: str = "configure") -> None:
        fn_path = module_name + "." + callback
        fn = import_string(fn_path)
        fn(self)

    load_from_callback = include

    def __iter__(self) -> t.Iterator[routing.BaseRoute]:
        return iter(self._routes)

    def __len__(self) -> int:
        return len(self._routes)

    @t.overload  # pragma: no cover
    def __getitem__(self, i: int) -> routing.BaseRoute:  # noqa: F811
        ...

    @t.overload  # pragma: no cover
    def __getitem__(self, s: slice) -> t.Sequence[routing.BaseRoute]:  # noqa: F811
        ...

    def __getitem__(  # noqa: F811
        self, index: t.Union[int, slice]
    ) -> t.Union[routing.BaseRoute, t.Sequence[routing.BaseRoute]]:
        """Not used, only here to match the interface of Sequence type.
        The router requires routes to be of Sequence type, this class
        mimics to the Sequence and therefore must implement __getitem__."""
        return self._routes[index]


class RouteURLResolver:
    def __init__(self, router: Router):
        self.router = router

    def resolve(self, name: str, **path_params: str) -> routing.URLPath:
        return self.router.url_path_for(name, **path_params)



from __future__ import annotations

import logging
import typing as t
from contextlib import AsyncExitStack

import click
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.container import Container

from .config import Config
from .contracts import Debug
from .dotenv import Env
from .errors import ErrorHandlers, ExceptionMiddleware
from .extensions import Extension, Extensions
from .middleware import MiddlewareStack
from .requests import Request
from .routing import Router, Routes


class App(Container):
    def __init__(
        self,
        config: dict = None,
        debug: bool = False,
        env_prefix: str = "",
        extensions: list[Extension] = None,
    ) -> None:
        super().__init__()
        self.commands: list[click.Command] = []
        self.config = Config(config)
        self.debug = debug
        self.env = Env(env_prefix)
        self.extensions = Extensions(extensions or [])
        self.error_handlers = ErrorHandlers()
        self.lifecycle: list[t.Callable[[App], t.AsyncContextManager]] = []
        self.middleware = MiddlewareStack()
        self.routes = Routes()

        self._asgi_app_instance: t.Optional[ASGIApp] = None
        self._booted = False

        self.bind(Debug, self.debug, aliases="debug")
        self.bind(App, self, "app")
        self.bind(Env, self.env, "env")
        self.bind(Config, self.config, aliases="config")
        self.bind(Routes, self.routes, aliases="routes")
        self.bind(MiddlewareStack, self.middleware, aliases="middleware")
        self.bind(Extensions, self.extensions, aliases="extensions")

    @property
    def asgi_app(self) -> ASGIApp:
        """Create ASGI application. Once created a cached instance returned."""
        try:
            if self._asgi_app_instance is None:
                self.bootstrap()
                app: ASGIApp = self.get(Router)
                self.middleware.top(
                    ExceptionMiddleware,
                    handlers=self.error_handlers,
                )
                self.middleware.top(ServerErrorMiddleware, debug=self.debug)
                for mw in reversed(self.middleware):
                    app = mw.wrap(app)
                self._asgi_app_instance = app
        except Exception as ex:
            logging.exception(ex)
            raise
        else:
            return self._asgi_app_instance

    def run_cli(self) -> None:
        """Run command line application."""
        self.bootstrap()
        app = click.Group()
        for command in self.commands:
            app.add_command(command)
        app()

    def bootstrap(self) -> None:
        if self._booted:
            return

        # register extension services
        for ext in self.extensions:
            ext.register(self)

        # bootstrap extension services
        for ext in self.extensions:
            ext.bootstrap(self)

    async def lifespan(self, app: App) -> t.AsyncGenerator:
        async with AsyncExitStack() as stack:
            for hook in self.lifecycle:
                await stack.enter_async_context(hook(app))
            yield

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        """ASGI entry point."""
        assert scope["type"] in {"http", "websocket", "lifespan"}

        scope["app"] = self
        if scope["type"] == "http":
            scope["request"] = Request(scope, receive, send)
        await self.asgi_app(scope, receive, send)



from .authentication import AuthenticationExtension
from .hashers import HashingExtension
from .jinja import JinjaExtension
from .routing import RoutingExtension

__all__ = [
    "JinjaExtension",
    "RoutingExtension",
    "AuthenticationExtension",
    "HashingExtension",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kupala',
 'kupala.framework',
 'kupala.framework.commands',
 'kupala.framework.extensions',
 'kupala.security']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0,<4.0',
 'aiofiles>=0.6.0,<0.7.0',
 'click>=7.1.2,<8.0.0',
 'itsdangerous>=1.1.0,<2.0.0',
 'passlib>=1.7.4,<2.0.0',
 'python-dotenv>=0.17.1,<0.18.0',
 'python-multipart>=0.0.5,<0.0.6',
 'starlette>=0.14,<0.15']

setup_kwargs = {
    'name': 'kupala',
    'version': '0.4.0',
    'description': '',
    'long_description': None,
    'author': 'alex.oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import pathlib
import typing as t

import jinja2

from kupala.application import App
from kupala.contracts import TemplateRenderer, URLResolver
from kupala.extensions import Extension
from kupala.security.csrf import CSRF_POST_FIELD, csrf_token

JinjaTemplateDirs = t.List[str]
JinjaFilters = t.Dict[str, t.Callable]
JinjaTests = t.Dict[str, t.Callable]
JinjaGlobals = t.Dict[str, t.Callable]
JinjaExtensions = t.List[str]


def csrf_input() -> str:
    return '<input name="%s" type="hidden" value="%s">' % (
        CSRF_POST_FIELD,
        csrf_token(),
    )


class JinjaRenderer(TemplateRenderer):
    """A template renderer that uses Jinja2 template engine."""

    def __init__(self, env: jinja2.Environment) -> None:
        self.env = env

    def render(
        self,
        template_name: str,
        context: dict[str, t.Any] = None,
    ) -> str:
        """Render a template."""
        template = self.env.get_template(template_name)
        return template.render(context or {})


class JinjaExtension(Extension):
    def __init__(
        self,
        template_dirs: list[t.Union[str, pathlib.Path]],
        loader: jinja2.loaders.BaseLoader = None,
        filters: dict = None,
        tests: dict = None,
        globals: dict = None,
        extensions: list[str] = None,
    ) -> None:
        self.loader = loader
        self.dirs = template_dirs
        self.filters = filters or {}
        self.tests = tests or {}
        self.globals = globals or {}
        self.extensions = extensions or []

    def register(self, app: App) -> None:
        app.bind(JinjaTemplateDirs, self.dirs, aliases="jinja.template_dirs")
        app.bind(JinjaFilters, self.filters, aliases="jinja.filters")
        app.bind(JinjaTests, self.tests, aliases="jinja.tests")
        app.bind(JinjaGlobals, self.globals, aliases="jinja.globals")
        app.bind(JinjaExtensions, self.extensions, aliases="jinja.extensions")

        def on_env_created(env: jinja2.Environment) -> None:
            env.filters.update(app.get(JinjaFilters))
            env.tests.update(app.get(JinjaTests))
            env.globals.update(app.get(JinjaGlobals))

        app.singleton(
            jinja2.Environment,
            self._jinja_factory,
            aliases="jinja",
        ).after_created(on_env_created)

        app.singleton(
            JinjaRenderer,
            self._jinja_renderer,
            aliases=["jinja.renderer", TemplateRenderer],
        )

    def _jinja_factory(self, app: App) -> jinja2.Environment:
        if self.loader is None:
            self.loader = jinja2.loaders.ChoiceLoader(
                [
                    jinja2.loaders.FileSystemLoader(list(reversed(self.dirs))),
                ]
            )

        env = jinja2.Environment(
            loader=self.loader,
            extensions=app.get(JinjaExtensions),
        )

        env.globals["url"] = app.get(URLResolver).resolve
        env.globals["csrf_input"] = csrf_input
        return env

    def _jinja_renderer(self, env: jinja2.Environment) -> JinjaRenderer:
        return JinjaRenderer(env)



import re
import typing as t

from starlette import datastructures as ds, requests

from kupala.sessions import Session

if t.TYPE_CHECKING:
    from kupala.authentication import AuthState

undefined = object()

Caster = t.Callable[[t.Any], t.Any]


class QueryParams(ds.QueryParams):
    def get_int(self, name: str, default: int = None) -> int:
        return self.get(name, default, int)

    def get_float(self, name: str, default: float = None) -> float:
        return self.get(name, default, float)

    def get_int_list(self, name: str) -> list[int]:
        return t.cast(list[int], self.getlist(name, int))

    def get_float_list(self, name: str) -> list[float]:
        return t.cast(list[float], self.getlist(name, float))

    def get_list(self, name: str) -> list:
        return self.getlist(name)

    def get(
        self,
        name: str,
        default: t.Any = None,
        cast: Caster = None,
    ) -> t.Any:
        value = super().get(name, default)
        if cast and value is not None:
            value = cast(value)
        return value

    def getlist(self, name: str, cast: Caster = None) -> list[str]:
        value = super().getlist(name)
        if cast:
            value = list(map(cast, value))
        return value


class Request(requests.Request):
    _query_params: QueryParams

    @property
    def auth(self) -> "AuthState":
        assert (
            "auth" in self.scope
        ), "AuthenticationMiddleware must be installed to access request.auth"
        return self.scope["auth"]

    @property
    def wants_json(self) -> bool:
        """Test if request sends Accept header
        with 'application/json' value."""
        return "application/json" in self.headers.get("accept", "")

    @property
    def is_ajax(self) -> bool:
        """
        Is true when the request is a XMLHttpRequest.
        It works if JavaScript's HTTP client sets an X-Requested-With HTTP header.

        Known frameworks:
        http://en.wikipedia.org/wiki/List_of_Ajax_frameworks#JavaScript
        """
        return self.headers.get("x-requested-with", None) == "XMLHttpRequest"

    @property
    def ip(self) -> str:
        """Returns the IP address of user."""
        return self.client.host

    @property
    def secure(self) -> bool:
        """Determine if the request served over SSL."""
        return self.scope["scheme"] == "https"

    @property
    def is_post(self) -> bool:
        """Test if request was made using POST method."""
        return self.method.upper() == "POST"

    @property
    def query_params(self) -> QueryParams:
        """Return query parameters."""
        if not hasattr(self, "_query_params"):
            self._query_params = QueryParams(self.scope.get("query_string", ""))
        return self._query_params

    @property
    def session(self) -> Session:  # type: ignore[override]
        assert (
            "session" in self.scope
        ), "SessionMiddleware must be installed to access request.session"
        return self.scope["session"]

    def old_data(self, default: t.Any = None) -> t.Any:
        data = None
        if "_redirect_data" in self.session:
            data = self.session.get("_redirect_data", default)
            del self.session["_redirect_data"]
        return data

    def url_matches(self, *patterns: str) -> bool:
        for pattern in patterns:
            if pattern == self.url.path:
                return True
            if re.match(pattern, self.url.path):
                return True
        return False

    def full_url_matches(self, *patterns: str) -> bool:
        for pattern in patterns:
            if pattern == str(self.url):
                return True
            if re.match(pattern, str(self.url)):
                return True
        return False



import abc
import typing as t

from passlib import hash
from passlib.ifc import PasswordHash


class BaseHasher(abc.ABC):  # pragma: nocover
    algorithm: t.Any = None

    def get_algorithm(self) -> PasswordHash:
        if self.algorithm is not None:
            return self.algorithm
        raise NotImplementedError()

    def hash(self, plain: str) -> str:
        return self.get_algorithm().hash(plain)

    def verify(self, plain: str, hashed: str) -> bool:
        return self.get_algorithm().verify(plain, hashed)

    def needs_update(self, hashed: str) -> bool:
        return self.get_algorithm().needs_update(hashed)


class Pbkdf2Sha256Hasher(BaseHasher):
    algorithm = hash.pbkdf2_sha256


class Pbkdf2Sha512Hasher(BaseHasher):
    algorithm = hash.pbkdf2_sha512


class InsecureHasher(BaseHasher):
    def hash(self, plain: str) -> str:
        return plain

    def verify(self, plain: str, hashed: str) -> bool:
        return plain == hashed

    def needs_update(self, hashed: str) -> bool:
        return False



from __future__ import annotations

import abc
import inspect
import typing as t
from contextlib import contextmanager

S = t.TypeVar("S", covariant=True)
N = t.Union[str, t.Type[S]]
Factory = t.Union[t.Callable, t.Type]
PostCreateHook = t.Callable[[t.Union[S, t.Any]], None]


class ContainerError(Exception):
    """Base exception for all container errors."""


class AliasExistsError(ContainerError):
    """Raised when alias name is already taken."""


class ServiceNotFound(ContainerError):
    """Raised when container fails to retrieve service."""


class ParameterUnsupported(ContainerError):
    """Raised if function or constructor parameter cannot be processed."""


class PositionOnlyArgumentError(ParameterUnsupported):
    """Raised if function or constructor parameter is of positional type."""


class NoTypeHintError(ContainerError):
    """Raised if function or constructor parameter has no type hint."""


class Binding:
    def __init__(self, service: N, container: Container):
        self._service = service
        self._container = container

    def alias(self, aliases: t.Union[N, list[N]]) -> Binding:
        """Alias an existing services with an alternate name."""
        self._container.alias(self._service, aliases)
        return self

    def tag(self, tags: t.Union[str, list[str]]) -> Binding:
        """Assign tags to a service."""
        self._container.tag(self._service, tags)
        return self

    def after_created(self, hook: PostCreateHook) -> Binding:
        """Call a hook after the service has been created by factory."""
        self._container.add_post_create_hook(self._service, hook)
        return self


class _ServiceGetter:
    def __init__(self, name: str):
        self.name = name


def service(name: str) -> _ServiceGetter:
    """Inject a dependency using a string service name."""
    return _ServiceGetter(name)


class ByTag:
    def __init__(self, tag: str):
        self.tag = tag


def tag(tag: str) -> ByTag:
    """Inject a list of dependencies for a specific tag."""
    return ByTag(tag)


class ResolveContext(abc.ABC):
    """Keeps resolved services."""

    def __init__(self) -> None:
        self._service_map: dict[N, BaseResolver] = {}

    def __setitem__(self, service: N, value: t.Any) -> None:
        self._service_map[service] = value

    def __getitem__(self, service: N) -> t.Any:
        return self._service_map[service]

    def __contains__(self, service: N) -> bool:
        return service in self._service_map

    def __enter__(self) -> ResolveContext:
        self._service_map = {}
        return self

    def __exit__(self, *args: t.Any) -> None:
        self._service_map.clear()
        self._service_map = {}


class BaseResolver(abc.ABC):  # pragma: nocover
    """Base resolver class."""

    def resolve(self, container: Container, context: ResolveContext) -> t.Any:
        raise NotImplementedError()


class InstanceResolver(BaseResolver):
    """Resolver that keeps an instance of the service."""

    def __init__(self, instance: t.Any) -> None:
        self.instance = instance

    def resolve(self, container: Container, context: ResolveContext) -> t.Any:
        return self.instance


class _HookDispatcher:
    def post_created(
        self,
        instance: t.Any,
        service: N,
        container: Container,
    ) -> None:
        callbacks = container.get_post_create_hooks(service)
        for callback in callbacks:
            callback(instance)


class FactoryResolver(BaseResolver, _HookDispatcher):
    """Resolver that creates a service from a factory function."""

    def __init__(self, name: N, factory: Factory) -> None:
        self.name = name
        self.factory = factory

    def resolve(self, container: Container, context: ResolveContext) -> t.Any:
        instance = container.invoke(self.factory)
        self.post_created(instance, self.name, container)
        return instance


class SingletonResolver(BaseResolver, _HookDispatcher):
    """Resolver that creates a singleton service from a factory function.
    If `scoped` is True then a created instance will be scoped
    to the resolve context."""

    def __init__(self, name: N, factory: Factory, scoped: bool) -> None:
        self.name = name
        self.factory = factory
        self.scoped = scoped
        self.instance = None

    def resolve(self, container: Container, context: ResolveContext) -> t.Any:
        if self.scoped:
            if self.name not in context:
                context[self.name] = container.invoke(self.factory)
                self.post_created(context[self.name], self.name, container)
            return context[self.name]

        if not self.instance:
            self.instance = container.invoke(self.factory)
            self.post_created(self.instance, self.name, container)
        return self.instance


class Container:
    """A container for registered service instances and factories."""

    def __init__(self) -> None:
        self._service_map: dict[N, BaseResolver] = {}
        self._aliases: dict[N, N] = {}
        self._tags: dict[str, list[N]] = {}
        self._post_create_hooks: dict[N, list[PostCreateHook]] = {}
        self._current_context = ResolveContext()

    @property
    def aliases(self) -> dict[N, N]:
        return self._aliases

    @property
    def tags(self) -> dict[str, list[N]]:
        return self._tags

    @property
    def post_create_hooks(self) -> dict[N, list[PostCreateHook]]:
        return self._post_create_hooks

    def get(self, name: N) -> t.Union[S, t.Any]:
        """Retrieve a service instance from the container."""
        real_name = self.resolve(name)
        resolver = self._service_map[real_name]
        instance = resolver.resolve(self, self._current_context)
        return t.cast(S, instance)

    def invoke(self, fn: t.Union[t.Callable, type], **args: t.Any) -> t.Any:
        """Invoke a callable or class constructor injecting dependencies."""
        signature = inspect.signature(fn)
        if inspect.isclass(fn):
            type_hints = t.get_type_hints(getattr(fn, "__init__"))
        else:
            type_hints = t.get_type_hints(fn)

        injections = args
        for p in signature.parameters.values():
            if p.kind == inspect.Parameter.POSITIONAL_ONLY:
                raise PositionOnlyArgumentError(
                    'Argument "%s" of "%s": '
                    "positional only arguments not supported. " % (p.name, fn)
                )

            if p.name not in injections:
                if p.name not in type_hints:
                    raise NoTypeHintError(
                        'Argument "%s" of "%s" has no type hint.' % (p.name, fn)
                    )

                dependency_type = type_hints[p.name]
                resolved_dep: t.Any
                if isinstance(dependency_type, _ServiceGetter):
                    resolved_dep = self.get(dependency_type.name)
                elif isinstance(dependency_type, ByTag):
                    resolved_dep = self.get_by_tag(dependency_type.tag)
                else:
                    resolved_dep = self.get(dependency_type)
                injections[p.name] = resolved_dep

        return fn(**injections)

    def bind(
        self,
        name: N,
        instance: t.Any,
        aliases: t.Union[N, list[N]] = None,
        tags: t.Union[str, list[str]] = None,
    ) -> Binding:
        """Bind an instance to the container.
        A cached instance always returned."""
        self._service_map[name] = InstanceResolver(instance)
        if aliases:
            self.alias(name, aliases)
        if tags:
            self.tag(name, tags)
        return Binding(name, self)

    def factory(
        self,
        name: N,
        factory: Factory,
        singleton: bool = False,
        aliases: t.Union[N, list[N]] = None,
        tags: t.Union[str, list[str]] = None,
        scoped: bool = False,
    ) -> Binding:
        """Bind a services factory.
        The service factory is a callable that creates a service instance.
        By default, it always creates a fresh instance unless `singleton`
        is specified."""
        if singleton:
            self._service_map[name] = SingletonResolver(name, factory, scoped)
        else:
            self._service_map[name] = FactoryResolver(name, factory)

        if aliases:
            self.alias(name, aliases)
        if tags:
            self.tag(name, tags)
        return Binding(name, self)

    def singleton(
        self,
        name: N,
        factory: Factory,
        aliases: t.Union[N, list[N]] = None,
        tags: t.Union[str, list[str]] = None,
        scoped: bool = False,
    ) -> Binding:
        """Bind a factory for a singleton service.
        Each time you ask for a service, a cached instance will be returned."""
        return self.factory(
            name,
            factory,
            singleton=True,
            aliases=aliases,
            tags=tags,
            scoped=scoped,
        )

    def alias(self, service: N, aliases: t.Union[N, list[N]]) -> None:
        """Alias an existing services with an alternate name."""
        aliases = [aliases] if isinstance(aliases, (str, type)) else aliases
        for alias_ in aliases:
            if alias_ in self._aliases:
                raise AliasExistsError(
                    'Alias %s is taken by "%s" service.'
                    % (
                        alias_,
                        self._aliases[alias_],
                    )
                )

            real_name = self.resolve(service)
            self._aliases[alias_] = real_name

    def get_aliases(self, service: N) -> list[t.Union[str, type]]:
        """Get all aliases assigned to the service."""
        return [
            alias for alias, real_name in self._aliases.items() if real_name == service
        ]

    def tag(self, service: N, tags: t.Union[str, list[str]]) -> None:
        """Assign tags to the service."""
        tags = [tags] if isinstance(tags, str) else tags
        for tag in tags:
            self._tags.setdefault(tag, [])
            self._tags[tag].append(service)

    def get_by_tag(self, tag: str) -> list[t.Any]:
        """Get services tagged with `tag`."""
        return list(map(self.get, self._tags.get(tag, [])))

    def get_tags(self, service: N) -> list[str]:
        """Get all tags assigned to the service."""
        return [tag for tag, services in self._tags.items() if service in services]

    def has(self, name: N) -> bool:
        """Test if container contains service `name`."""
        try:
            return self.resolve(name) in self._service_map
        except ServiceNotFound:
            return False

    def add_post_create_hook(self, name: N, hook: PostCreateHook) -> None:
        """Add a callable to execute after successful service creation."""
        self._post_create_hooks.setdefault(name, [])
        self._post_create_hooks[name].append(hook)

    def get_post_create_hooks(self, name: N) -> list[PostCreateHook]:
        return self._post_create_hooks.get(name, [])

    def resolve(self, name: N) -> N:
        """Resolve a service name or an alias into an actual service name."""
        if name in self._aliases:
            return self.resolve(self._aliases[name])

        if name not in self._service_map:
            raise ServiceNotFound('Name "%s" not found in the container.' % name)
        return name

    def remove(self, name: N) -> None:
        """Remove service from the container.
        This will also remove all aliases and tags."""
        if name in self._service_map:
            del self._service_map[name]

        for tag in self.get_tags(name):
            self._tags[tag].remove(name)
            if len(self._tags[tag]) == 0:
                del self._tags[tag]

        for alias in self.get_aliases(name):
            del self._aliases[alias]

        if name in self._post_create_hooks:
            del self._post_create_hooks[name]

    @contextmanager
    def scoped(self, context: ResolveContext = None) -> t.Iterator[None]:
        context = context or ResolveContext()
        self._current_context = context
        yield
        self._current_context = ResolveContext()

    __contains__ = has
    __getitem__ = get
    __setitem__ = bind
    __delitem__ = remove
